(* TZoomImage A tImage derivative that supports zooming.

This component is postcardware: if you like it and use it,
send me a postcard at

Grahame Grieve
Biochemistry
St Vincent's Hospital
Fitzroy 3065
Australia

Disclaimer: I'm not responsible for any mishaps that result
from the use of this component/code.

This code is copyright  Grahame Grieve,
   g.grieve@pgrad.unimelb.edu.au

Description: this component shows a bitmap, like TImage,
but you can control the zooming of the bitmap. 

The component is for both versions of Delphi.

Version 1.1: Automatic mouse zooming. 5/8/1996

bugs: In Delphi32 some images (b/w) don't load properly. I don't
know why (p.s. if you find out, email me with the answer).

>>>>>>  How to use this component:  <<<<<<<<<<<

how to set zoom:
  1. you can set autofit to fit the picture. When autofit is true
     the other properties become effectively read-only. but setting them
     sets autofit to false
  2. You can use zoomvalue and xcentre,ycentre
     if you set any of these values, the others are held constant,
     and the visXXX values all change. if the x and y zooms are different,
     zoomvalue returns the average of the 2 values. When you set ZoomValue,
     both are set to the same value.
     ZoomValue is a percent of picture size on a bit for bit basis. At
     100% there is a 1 to 1 bit copy.
  3. You can use vistop - visright.
     if you set one of these, the others remain constant, and the values
     for zoomvalue and y/x centre change.
  4. New!! the image automatically responds to a mouse drag by zooming
     to the are selected. single click clears the zooming.

Rules for picture sizing:
   * when there is no loaded picture, these values cannot be changed.
   * if the bitmap is changed, then it will be automatically scaled to fit.
   * You can refit the picture at any time by calling autofit, or setting
      zoomValue to 0.
   * at no time can the visible area of the bitmap become smaller
     than the actual size of the client rect. i.e. Requests to set visXXX
     outside the bounds of the bitmap result in the value of visXXX being
     scaled to the limit.
   * abs(visRight - VisLeft) must > mindiff
     abs(VisBottom - VisTop) must > minDiff  *)


unit Zoompic;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, menus, extctrls;

const minDiff = 10; (* controls the zoom limit of the bitmap *)

type
  TZoomImage = class(TGraphicControl)
  private
    Fvistop,Fvisleft,Fvisbottom,Fvisright:word;
    FinVert,fdragging:boolean;
    dx,dy:integer;
    drect:trect;

    procedure PictureChanged(Sender: TObject);
    procedure SetBitmap(Value: TBitmap);
    function hasbitmap:boolean;
    procedure setvistop(v:word);
    procedure setinvert(v:boolean);
    procedure setvisleft(v:word);
    procedure setvisbottom(v:word);
    procedure setvisright(v:word);
    function getZoomV:word;
    procedure setZoomV(v:word);
    function getXCenter:word;
    function getYCenter:word;
    procedure setXCenter(v:word);
    procedure setYCenter(v:word);
  protected
    FBitmap: TBitmap;
    FStretchmode:integer; {so's descendants can control the painting mode }
    procedure Paint; override;
    {Implements dragging and resizing:}
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AutoFit;
  published
    property zoomValue:word read getZoomV write setZoomV;
    property XCenter:word read getXCenter write SetXCenter;
    property YCenter:word read getYCenter write SetYCenter;
    property vistop:word read Fvistop write setvistop;
    property visleft:word read Fvisleft write setvisleft;
    property visbottom:word read Fvisbottom write setvisbottom;
    property visright:word read Fvisright write setvisright;
    property Invert:boolean read Finvert write setinvert;
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ParentShowHint;
    property Bitmap: TBitmap read FBitmap write SetBitmap;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

uses Consts;

procedure Register;
begin
  RegisterComponents('Image', [TZoomImage]);
end;

constructor TZoomImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBitmap := TBitmap.Create;
  FBitmap.OnChange := PictureChanged;
  Height := 160;
  Width := 160;
  Fvistop := 0;
  Fvisleft := 0;
  Fvisbottom := mindiff;
  Fvisright := mindiff;
  FStretchmode := STRETCH_ANDSCANS;
  drect := rect(0,0,0,0);
end;

destructor TZoomImage.Destroy;
begin
  FBitmap.Free;
  inherited Destroy;
end;

function TZoomImage.hasBitmap;
begin
  result := (FBitmap.Handle <> 0);
end;

procedure TZoomImage.setvistop(v:word);
begin
  if hasBitMap and (v <> FVisTop) then
   begin
     if v < 0 then FVisTop := 0 else FVisTop := v;
     if FVisBottom - FVisTop < MinDiff then FVisTop := FVisBottom - MinDiff;
     invalidate;
   end;
end;

procedure TZoomImage.setvisleft(v:word);
begin
  if hasBitMap and (v <> FVisLeft) then
   begin
     if v < 0 then FVisLeft := 0 else FVisLeft := v;
     if FVisRight - FVisLeft < MinDiff then FVisLeft := FVisRight - MinDiff;
     invalidate;
   end;
end;

procedure TZoomImage.setvisbottom(v:word);
begin
  if hasBitMap and (v <> FVisBottom) then
   begin
     if v > FBitmap.Height then FVisBottom := FBitmap.Height
                else FVisBottom := v;
     if FVisBottom - FVisTop < MinDiff then FVisBottom  := FVisTop + MinDiff;
     invalidate;
   end;
end;

procedure TZoomImage.setvisright(v:word);
begin
  if hasBitMap and (v <> FVisRight) then
   begin
     if v > FBitmap.Width then FVisRight := FBitmap.Width
                else FVisRight := v;
     if FVisRight - FVisLeft < MinDiff then FVisRight := FVisLeft + MinDiff;
     invalidate;
   end;
end;

function TZoomImage.getZoomV:word;
begin
 if hasbitmap then
     result := trunc((((width / (FVisRight - FVisLeft)) +
          ( Height / (FVisBottom - FVisTop))) / 2) * 100)
   else
     result := 100;
end;

procedure TZoomImage.setZoomV(v:word);
var cx,cy,nw,nh:word;
begin
 if hasbitmap and (v >= 0) then
  if v = 0 then autofit else
   begin
    cx := getXCenter;
    nw := trunc(width / (v/100));
    if nw > FBitMap.width then nw := FBitMap.width;
    if nw < mindiff then nw := mindiff;
    if (cx - (nw div 2)) < 0 then cx := nw div 2;
    if (cx + (nw div 2)) > Fbitmap.width then cx := Fbitmap.width - (nw div 2);
    FVisLeft := cx - nw div 2;
    FVisRight := cx + nw div 2;
    cy := getyCenter;
    nh := trunc(height / (v/100));
    if nh < mindiff then nh := mindiff;
    if nh > FBitMap.height then nh := Fbitmap.height;
    if (cy - (nh div 2)) < 0 then cy := nh div 2;
    if (cy + (nh div 2)) > Fbitmap.height then cy := Fbitmap.width - (nh div 2);
    FVisTop := cy - nh div 2;
    FVisBottom := cy + nh div 2;
    invalidate;
   end;
end;

function TZoomImage.getXCenter:word;
begin
 result := (FVisLeft + FVisRight) div 2;
end;

function TZoomImage.getYCenter:word;
begin
 result := (FVisTop + FVisBottom) div 2;
end;

Procedure TZoomImage.setXCenter(v:word);
var sleft,sright,shift:integer;
begin
 if hasbitmap then
  begin
   shift := v-GetXCenter;
   sleft := FVisLeft;
   sRight := FBitmap.width - FVisRight;
   if shift > sRight then shift := SRight;
   if shift < 0 - sLeft then shift := 0 - SLeft;
   FVisLeft := FVisLeft + shift;
   FVisRight := FVisRight + shift;
   invalidate;
  end;
end;

procedure TZoomImage.SetInvert;
begin
 if v <> Finvert then
  begin
   FinVert := v;
   if v then FStretchmode := Stretch_orscans else FStretchmode := stretch_andscans;
   invalidate;
  end;
end;

Procedure TZoomImage.setYCenter(v:word);
var stop,sbottom,shift:integer;
begin
 if hasbitmap then
  begin
   shift := v-GetYCenter;
   stop := FVisTop;
   sbottom := FBitmap.height - FVisBottom;
   if shift > sbottom then shift := Sbottom;
   if shift < 0 - stop then shift := 0 - Stop;
   FVistop := FVistop + shift;
   FVisbottom := FVisbottom + shift;
   invalidate;
  end;
end;

procedure TZoomImage.AutoFit;
begin
  if hasbitmap then
   begin
     FVisTop := 0;
     FVisLeft := 0;
     FVisBottom := FBitmap.Height;
     FVisRight := FBitmap.Width;
     invalidate;
   end;
end;

procedure TZoomImage.Paint;
var
  Dest: TRect;
begin
  if csDesigning in ComponentState then
    with inherited Canvas do
    begin
      Pen.Style := psDash;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;
 if hasBitmap then
  begin
   Dest := ClientRect;
   SetStretchBltMode(inherited Canvas.Handle, FStretchmode);
   if not FinVert then
    StretchBlt(inherited Canvas.Handle, Dest.Left, Dest.Top,
          Dest.Right - Dest.Left, Dest.Bottom - Dest.Top,
      FBitmap.Canvas.Handle, FVisLeft, FVisTop, FVisRight - FVisLeft,
          FVisBottom - FVisTop, srccopy) else
    StretchBlt(inherited Canvas.Handle, Dest.Left, Dest.Top,
          Dest.Right - Dest.Left, Dest.Bottom - Dest.Top,
      FBitmap.Canvas.Handle, FVisLeft, FVisTop, FVisRight - FVisLeft,
          FVisBottom - FVisTop, notSrcCopy);
  end;
end;

procedure TZoomImage.SetBitmap;
begin
  FBitmap.Assign(Value);
end;

procedure TZoomImage.PictureChanged(Sender: TObject);
begin
  if (FBitmap.Width = Width) and
    (FBitmap.Height = Height) then
    ControlStyle := ControlStyle + [csOpaque] else
    ControlStyle := ControlStyle - [csOpaque];
  Fvistop := 0;
  Fvisleft := 0;
  if hasbitmap then
    begin
     Fvisbottom := FBitmap.Height;
     Fvisright := FBitmap.Width;
    end
   else
    begin
     Fvisbottom := mindiff;
     Fvisright := mindiff;
    end;
  Invalidate;
end;

procedure TZoomImage.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) then
    begin
    fdragging := true;
    dx := x;
    dy := y;
    screen.cursor := crCross;
    end;
  inherited MouseDown(Button, Shift, X, Y);
end {MouseDown};

procedure TZoomImage.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
 if fdragging then
 begin
   Canvas.DrawFocusRect(drect);
   if x > dx then
   begin
     drect.left := dx;
     drect.right := x;
   end
   else
   begin
     drect.left := x;
     drect.right := dx;
   end;
   if y > dy then
   begin
     drect.top := dy;
     drect.bottom := y;
   end
   else
   begin
     drect.top := y;
     drect.bottom := dy;
   end;
   Canvas.DrawFocusRect(drect);
 end;
 inherited mousemove(shift, x, y);
end;

procedure TZoomImage.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
const MouseTolerance = 4; {require mouse to move at least 4 pixels in both directions
       before counting as really moving the mouse - help clumsy mouse users!}
procedure swap(var x1,x2:integer);
var t:integer;
begin
  t := x1;
  x1 := x2;
  x2 := t;
end;

begin
  if Fdragging then
    begin
    drect := rect(0,0,0,0);
    releaseCapture;
    FDragging := False;
    screen.cursor := crDefault;
    if x > dx then swap(x,dx);
    if y > dy then swap(y,dy);
    if (dx - x > MouseTolerance) and (dy - y > MouseTolerance) then
      begin
      x := Fvisleft + trunc((FvisRight - FVisLeft) * (x / width));
      dx := Fvisleft + trunc((FvisRight - FVisLeft) * (dx / width));
      y := FVisTop + trunc((FVisBottom - FVisTop) * (y / height));
      dy := FVisTop + trunc((FVisBottom - FVisTop) * (dy / height));
      VisLeft := x;
      VisRight := dx;
      VisTop := y;
      VisBottom := dy;
      invalidate;
      end
    else autofit;
  end;
  inherited mouseup(button, shift, x, y);
end;



end.
